@@ -4,6 +4,7 @@
 #include <unistd.h>
 #include <sys/types.h>
 #include <sys/signal.h>
+#include <pthread.h>
 
 # ifndef O_CLOEXEC
 #  define O_CLOEXEC 02000000
@@ -30,6 +31,7 @@ typedef struct processx_handle_s {
 char *processx__tmp_string(SEXP str, int i);
 char **processx__tmp_character(SEXP chr);
 
+extern pthread_t processx__main_thread;
 void processx__sigchld_callback(int sig, siginfo_t *info, void *ctx);
 void processx__setup_sigchld();
 void processx__remove_sigchld();
@@ -71,7 +73,7 @@ void processx__create_control_write(processx_handle_t *handle,
 int processx__interruptible_poll(struct pollfd fds[],
 				 nfds_t nfds, int timeout);
 
-void processx__make_socketpair(int pipe[2]);
+void processx__make_socketpair(int pipe[2], const char *name);
 
 double processx__create_time(long pid);
 
